from __future__ import annotations
from typing import Union

from .feed_forward_neural_network import FeedForwardNeuralNetwork

class FFNNModelFile:
    def __init__(self, model_file: str) -> None:
        self.model_file: str = model_file
        self.load_file()

    def load_file(self) -> FFNNModelFile:
        # Load from self.model_file
        with open(self.model_file, "r") as model_file:
            model_txt: str = model_file.read()
            model_split = model_txt.split("\n\n")
            self.ffnn = FeedForwardNeuralNetwork(int(model_split[0]))
            for line in model_split[1:]:
                arg_arr = line.splitlines()
                act_func = arg_arr[0]
                totalNeurons = int(arg_arr[1])
                weight_dim = [int(dim) for dim in arg_arr[2].split()]
                weight = [float(w) for w in arg_arr[3].split()]
                self.ffnn.add_layer(
                        act_func,
                        totalNeurons,
                        weight_dim,
                        weight
                    )
        return self

    def get_ffnn(self) -> FeedForwardNeuralNetwork:
        return self.ffnn

if __name__ == "__main__":
    mf = FFNNModelFile("test/s2.txt")
    mf.load_file()
    for layer in mf.get_ffnn().layers:
        print(layer.activation_function)
        print(layer.total_neurons)
        print(layer.weight)



from __future__ import annotations

class Synapse:
    def __init__(
            self, 
            neuron_from: Neuron, 
            neuron_to: Neuron, 
            weight: float = 0,
            synapse_name: str = ""
        ) -> None:
        self.name : str = synapse_name
        self.weight : float = weight
        self.neuron_from : Neuron = neuron_from
        self.neuron_to : Neuron = neuron_to



from __future__ import annotations

from abc import ABC, abstractmethod


class ActivationFunction(ABC):
    @abstractmethod
    def setup(self, **kwargs) -> None:
        pass

    @abstractmethod
    def __call__(self, weighted_input):
        pass



from __future__ import annotations
from typing import List

from .layer import Layer

class FeedForwardNeuralNetwork:
    def __init__(self, totalLayers: int) -> None:
        self.totalLayers = totalLayers
        self.layers: List[Layer] = []
    
    def add_layer(
            self, 
            activation_function: str, 
            totalNeurons: int, 
            weight_dim: List[int], 
            weight: List[float]
        ):
        self.layers.append(
                Layer(
                    activation_function,
                    totalNeurons,
                    weight_dim,
                    weight
                )
            )



from __future__ import annotations
from typing import Callable, List

from .activation_function import softMax

class Layer:
    def __init__(
            self, 
            activation_function: str,
            totalNeurons: int,
            weight_dim: List[int],
            weight: List[float]
            ) -> None:
        self.neurons: List[Neuron] = []
        self.activation_function = activation_function
        self.total_neurons = totalNeurons
        self.weight_dim = weight_dim
        self.weight = weight

    def get_activation_function(self) -> Callable[..., float]:
        def act_func(f):
            return softMax(f, self.get_sum_of_weighted_sum())
        return act_func

    def get_sum_of_weighted_sum(self):
        sum_ = 0
        for neuron in self.neurons:
            sum_ += neuron.get_weighted_sum()
        return sum_



from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='malifpy_ffnn',
    version='0.0.4',
    license='MIT',
    author="Muhammad Alif Putra Yasa",
    author_email='malifputrayasa@gmail.com',
    packages=find_packages('./'),
    package_dir={'': './'},
    url='https://github.com/malifpy/malifpy_ffnn',
    keywords='ffnn',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'typing',
    ],
)



from __future__ import annotations
from typing import List, Union
from .synapse import Synapse

class Neuron:
    def __init__(self) -> None:
        self.name : str = ""
        self.output : Union[float, None] = None
        self.weighted_sum: Union[float, None] = None
        self.synapse_in  : List[Synapse] = []
        self.synapse_out : List[Synapse] = []
        self.layer : Layer

    def get_output(self) -> float:
        if self.output == None:
            self.output = self.layer.get_activation_function()(self.get_weighted_sum())
        return self.output

    def get_weighted_sum(self) -> float:
        if self.weighted_sum == None:
            self.weighted_sum = 0
            for synapse in self.synapse_in:
                self.weighted_sum += synapse.neuron_from.get_output() * synapse.weight
        return self.weighted_sum

    def connect_from(self, neuron_arg: Neuron) -> None:
        """
        Menghubungkan Neuron argumen **KE** Neuron ini.
        """
        new_synapse : Synapse = Synapse(neuron_arg, self)
        self.add_synapse_in(new_synapse)
        neuron_arg.add_synapse_out(new_synapse)

    def add_synapse_in(self, s: Synapse) -> None:
        """
        Memasukkan malifpy_ffnn.synapse.Synapse ke synapse_in
        """
        self.synapse_in.append(s)

    def connect_to(self, neuron_arg: Neuron) -> None:
        """
        Menghubungkan Neuron ini ke **KE** Neuron argumen.
        """
        new_synapse : Synapse = Synapse(neuron_arg, self)
        self.add_synapse_in(new_synapse)
        neuron_arg.add_synapse_out(new_synapse)

    def add_synapse_out(self, s: Synapse) -> None:
        """
        Memasukkan malifpy_ffnn.synapse.Synapse ke synapse_out
        """
        self.synapse_out.append(s)



from __future__ import annotations
from .activation_function import ActivationFunction

def softMax(weighted_input, sum_of_weighted_input):
    return weighted_input / sum_of_weighted_input

class SoftMax(ActivationFunction):
    def __init__(self) -> None:
        self.sum_of_weighted_sum: float

    def setup(self, **kwargs) -> SoftMax:
        self.sum_of_weighted_sum = kwargs["sum_of_weighted_sum"]
        return self

    def __call__(self, weighted_input):
        return weighted_input / self.sum_of_weighted_sum

